#include <fxsaber\ThirdPartyTicks\File.mqh> // https://www.mql5.com/ru/code/20225
#include <fxsaber\TesterCache\TesterCache.mqh> // https://www.mql5.com/ru/code/26223
#include <fxsaber\SingleTesterCache\SingleTesterCache.mqh> // https://www.mql5.com/ru/code/27611

#include "ValidateTask.mqh"
#include "ValidateBalance.mqh"

#define DONE "Done\\"

class VALIDATEEXE : public VALIDATEBASE
{
private:
  const string FolderName;
  string FileNames[];

  VALIDATETASK Task;
  VALIDATEBALANCE Balance;

  long HashOpt;
  long HashTst;

  bool FileDone( const int Num = 0 ) const
  {
    return(::FileMove(this.FileNames[Num], 0,
                      this.FolderName + DONE + ::StringSubstr(this.FileNames[Num], ::StringLen(this.FolderName)), FILE_REWRITE));
  }

  int Load( void )
  {
    ::ArrayFree(this.FileNames);

    if (this.FolderName != NULL)
    {
      const string Done = this.FolderName + DONE;
      const int DoneLen = ::StringLen(Done);

      for (int i = FILE::GetFileNames(this.FileNames, this.FolderName + "*.ini") - 1; i >= 0; i--)
        if (::StringSubstr(this.FileNames[i], 0, DoneLen) == Done)
          ::ArrayRemove(this.FileNames, i, 1);

      if (this.GetAmount())
        this.Task.Load(this.FileNames[0]);

      if (this.Balance.Load(this.FolderName))
        this.Balance.Visual();
    }

    return(this.GetAmount());
  }

  int GetAmount() const
  {
    return(::ArraySize(this.FileNames));
  }

  string GetFileName( void ) const
  {
    return(this.GetAmount() ? this.FileNames[0] : NULL);
  }

  static long GetHash( const uchar &Bytes[] )
  {
    long Res = 0;

    for (int i = ::ArraySize(Bytes) - 1; i >= 0; i--)
      Res += Bytes[i];

    return(Res);
  }

  static int GetMaxFitnessPos( const TESTERCACHE<ExpTradeSummary> &Cache, int Criterion = -1 )
  {
    int Pos = 0;
    double MaxFitness = -DBL_MAX;

    if (Criterion < 0)
      Criterion = Cache.Header.last_criterion;

    for (int i = Cache.GetAmount() - 1; i >= 0; i--)
    {
      const double Res = Cache[i].GetCriterionResult(Criterion);

      if (Res > MaxFitness)
      {
        MaxFitness = Res;

        Pos = i;
      }
    }

    return(Pos);
  }

  void WorkWithOptCache( void )
  {
    uchar Bytes[];

    if (MTTESTER::GetLastOptCache(Bytes))
    {
      const long Hash = VALIDATEEXE::GetHash(Bytes);

      if (Hash != this.HashOpt)
      {
        this.HashOpt = Hash;

        string FileName = this.GetFileName();
        FileName = ::StringSubstr(FileName, 0, ::StringLen(FileName) - ::StringLen("ini"));

        ::FileSave(this.FolderName + "Opt\\" + ::StringSubstr(FileName, ::StringLen(this.FolderName)) + "opt", Bytes);

        const TESTERCACHE<ExpTradeSummary> TesterCache(Bytes);
        ::ArrayFree(Bytes);

        const int Pos = VALIDATEEXE::GetMaxFitnessPos(TesterCache);

        TesterCache.SaveSet(Pos, this.FolderName + "Set\\" + ::StringSubstr(FileName, ::StringLen(this.FolderName)) + "set",
                            true, "Created by " + __FILE__);

        string Settings = this.Task.GetOutSampleSettings();

        if (Settings != NULL)
        {
          string StrTmp = Settings;
          const string StrOOS = VALIDATEEXE::StringBetween(StrTmp, "; OutSample: ", "\n");

          StrTmp = Settings;
          Settings = VALIDATEEXE::StringBetween(StrTmp, NULL, "[TesterInputs]");

          MqlParam Params[][5];

          for (int i = TesterCache.GetInputs(Pos, Params) - 1; i >= 0; i--)
            VALIDATEEXE::StringReplace(StrTmp, Params[i][PARAM_NAME].string_value + "=", "||", Params[i][PARAM_VALUE].string_value);

          Settings += "Visual=0\r\n[TesterInputs]\r\n" + StrTmp;

          if (this.Balance.GetAmount())
            VALIDATEEXE::StringReplace(Settings, "Deposit=", "\r\n", ::DoubleToString(this.Balance.GetLastBalance(), 2));

          FileName = this.FolderName + ::StringSubstr(FileName, ::StringLen(this.FolderName), ::StringLen("xxx_")) + StrOOS + "_tst.ini";

          ::StringToCharArray(Settings, Bytes);
          ::FileSave(FileName, Bytes);
        }
      }
    }

    return;
  }

  void WorkWithTstCache( void )
  {
    uchar Bytes[];

    if (MTTESTER::GetLastTstCache(Bytes))
    {
      const long Hash = VALIDATEEXE::GetHash(Bytes);

      if (Hash != this.HashTst)
      {
        this.HashTst = Hash;

        string FileName = this.GetFileName();
        FileName = ::StringSubstr(FileName, 0, ::StringLen(FileName) - ::StringLen("ini"));

        ::FileSave(this.FolderName + "Tst\\" + ::StringSubstr(FileName, ::StringLen(this.FolderName)) + "tst", Bytes);

        const SINGLETESTERCACHE SingleTesterCache(Bytes);

        SingleTesterCache.SaveSet(this.FolderName + "Set\\" + ::StringSubstr(FileName, ::StringLen(this.FolderName)) + "set",
                                  true, "Created by " + __FILE__);

        if (::ArraySize(SingleTesterCache.Deals) == 1)
          ::Alert("Initial balance (" + ::DoubleToString(SingleTesterCache.Header.trade_deposit, 2) + ") may be too small.");
        else if (this.Balance.Load(this.FolderName))
          this.Balance.SaveVisual(this.FolderName + "Img\\" + ::StringSubstr(FileName, ::StringLen(this.FolderName)) + "gif");
      }
    }

    return;
  }

  void WorkWithCache( void )
  {
    if (this.Task.IsOptimization())
    {
      MTTESTER::CloseNotChart();

      this.WorkWithOptCache();
    }
    else
      this.WorkWithTstCache();

    return;
  }

public:
  VALIDATEEXE( const string sFolderName, const string sOutSampleSymbol = "" ) : FolderName(sFolderName), HashOpt(0), HashTst(0), Task(sOutSampleSymbol)
  {
    if (!::MQLInfoInteger(MQL_TESTER) && this.Load())
    {
      uchar Bytes[];

      if (MTTESTER::GetLastOptCache(Bytes))
      {
        this.HashOpt = VALIDATEEXE::GetHash(Bytes);

        ::ArrayFree(Bytes);
      }

      if (MTTESTER::GetLastTstCache(Bytes))
        this.HashTst = VALIDATEEXE::GetHash(Bytes);
    }
  }

  bool IsFinish( void ) const
  {
    return(!this.GetAmount());
  }

  void OnTimer( void )
  {
    if (!::IsStopped() && !this.IsFinish())
    {
      if (this.Task.IsFinish())
      {
        ::Alert(this.GetFileName() + " - Done. " + (string)(this.GetAmount() - 1) + " tasks left.\n");

        this.WorkWithCache();

        this.FileDone();

        this.Load();
      }

      if (this.Task.Run())
        ::Alert(this.GetFileName() + " - Run. " + (string)this.GetAmount() + " tasks left.");
    }
  }
};

#undef DONE